-- collect CC data and apply filter to correct PV generated power overlayed on top of mains.

module(...,package.seeall)

require("xap")
require("xap.bsc")
require("string")
require("pl")

info={
   version="1.0 BETA", description="Adjusted Power Consumption", username="PEEKS"
}

-- SET UP THE VARIABLES

local GenPower = 1
local MainsPower = 1
local BoilerPower = 1
local TempGenerationW = 0
local TempMainsW = 0
local TempBoilerW = 0
local GenerationW = 0
local MainsW = 0
local BoilerW = 0
local CurrentGenW = 0
local CurrentMainsW = 0
local CurrentBoilerW = 0
local CumulativeW = 0
local ConsumedW = 0
local BaseW = 1
local QUAD = 0
local EST = 0
local GEN = 0
local DebugStatus = 0
local hysteresis = 5
local AverageW = 0
local Post = 1

-- START THE APP.............

function init()

-- Set up a new End Point
	bscConsumedW = bsc.Endpoint{source="dbzoo.livebox.HAH11.plugboard:ConsumedW", direction=bsc.INPUT, type=bsc.STREAM}

-- Get the previously stored Consumption value
	local file = io.open("/etc/plugboard/ConsumedW","r")
	ConsumedW= tonumber(file:read("*all"))
	file:close()

-- And the Base level.  THIS IS THE 'FLOOR' LEVEL FOR YOUR HOUSE - YOU'LL HAVE TO MAKE AN EDUCATED GUESS
--                      BASED ON WHAT READINGS YOU GET (AFTER MIDNIGHT) ON CurrentCost:chX(?)

	local file = io.open("/etc/plugboard/BaseW","r")
	BaseW = tonumber(file:read("*all"))
	file:close()

-- And the Running Average Consumption level - based on a hysteresis of 5 - 20 (ish - higher number = slower changes!!)
	local file = io.open("/etc/plugboard/AverageW","r")
	AverageW = tonumber(file:read("*all"))
	file:close()

-- This is where we get the quadratic equation values for your setup....they are in a table called Q

	Q={q1=0,q2=0,q3=0,q4=0,q5=0}
--	Q=pretty.read(file.read("/etc/plugboard/QUADvalue"))

-- Check the initialisation level
	if ConsumedW < BaseW then
		ConsumedW = BaseW
	end
-- and publish it
	bscConsumedW:setText( tostring(ConsumedW) )
	bscConsumedW:sendEvent()


--*******************************************************************************************************************DATA READ
--*******************************************************************************************************************
-- Get the data *********************  vvvvvv CHANGE CHANNELS TO SUIT YOUR SETUP  vvvvv

-- GENERATED POWER FROM THE SOLAR PV
  	local fa = xap.Filter()
  	fa:add("xap-header","source","dbzoo.livebox.HAH11.CurrentCost:ch.2")
  	fa:add("xap-header","class","xAPBSC.event")
	fa:callback(GenPowerLog)

-- CONSUMED POWER OFF THE MAINS
  	local fb = xap.Filter()
  	fb:add("xap-header","source","dbzoo.livebox.HAH11.CurrentCost:ch.1")
  	fb:add("xap-header","class","xAPBSC.event")
	fb:callback(MainsPowerLog)

-- POWER DUMPED INTO THE BOILER ************************* MUST BE ON ITS OWN POWER LINE !!!!
  	local fc = xap.Filter()
  	fc:add("xap-header","source","dbzoo.livebox.HAH11.CurrentCost:ch.3")
  	fc:add("xap-header","class","xAPBSC.event")
	fc:callback(BoilerPowerLog)

-- I've set my refresh to 6 secs..... change to suit
xap.Timer(filterW, 6):start()

end




--*******************************************************************************************************************FILTER FUNCTION
--*******************************************************************************************************************
-- THIS IS THE FILTER FUNTION

function filterW(self)

	GenerationW = tonumber(TempGenerationW)

-- SET THE SOLAR PV TO ZERO IF ITS BELOW THIS THRESHOLD ************************************************************** CHANGE TO SUIT
	if GenerationW < 68 then
		GenerationW = 0
	end

	MainsW = tonumber(TempMainsW)

	BoilerW = tonumber(TempBoilerW)


--*******************************************************************************************************************POLYNOMIAL
--*******************************************************************************************************************
-- POLYNOMIAL =  0.0000000000615*GEN^4 - 0.00000055*GEN^3 + 0.001805*GEN^2 - 1.565*GEN + 807.64

	GEN = GenerationW
	if GEN < 1050 then
		QUAD = Q.q1 * math.pow(GEN , 4)
		QUAD = QUAD - (Q.q2 * math.pow(GEN ,3))
		QUAD = QUAD + (Q.q3 * math.pow(GEN ,2))
		QUAD = QUAD - (Q.q4 * GEN ) + Q.q5
	else
		QUAD = (0.9709 * GEN) - 450.67
	end

	QUAD = math.floor(QUAD + 0.5)

	QUAD = 0

	if GenerationW > 0 then

		if MainsW > GenerationW then

			EST = MainsW + GenerationW - BaseW

			if BoilerW > 1 then

				EST = EST - MainsW

			end

		else

			EST = math.abs(MainsW - QUAD) + BaseW

		end

	else

		EST = MainsW

	end

	EST = MainsW
	EST = math.floor(EST + 0.5)
	CumulativeW = EST

-- Basic upper and lower limit checks
	if CumulativeW < tonumber(BaseW) then
		CumulativeW = math.floor(BaseW + (MainsW - (math.floor(MainsW/10)*10)))
	end
	if CumulativeW > tonumber(GenerationW + MainsW) or CumulativeW < 1 or CumulativeW == nil then
		CumulativeW = MainsW
	end

--*******************************************************************************************************************HYSTERESIS
--*******************************************************************************************************************
-- Do the moving average check and calc
	if tonumber(CumulativeW) < 1150 then
		if tonumber(CumulativeW) > AverageW then
			CumulativeW =  math.ceil(tonumber(((AverageW * hysteresis) + tonumber(CumulativeW)) / (hysteresis + 1)))
		else
			CumulativeW =  math.ceil(tonumber((AverageW + tonumber(CumulativeW)) / 2))
		end
	end

	ConsumedW= math.ceil(tonumber(CumulativeW))
	if ConsumedW < BaseW then
		ConsumedW = BaseW
	end
	bscConsumedW:setText( tostring(ConsumedW) )
	bscConsumedW:sendEvent()
	AverageW = ConsumedW


--*******************************************************************************************************************POST DATA TO FILE
--*******************************************************************************************************************
-- I've set the Post to 9 to reduce the amount of data collected..... lower the number to collect more, if needed....

	if Post > 9 then
		local file = io.open("/etc/plugboard/ConsumedW","w+")
		file:write(tostring(ConsumedW))
		file:close()

		local file = io.open("/etc/plugboard/AverageW","w+")
		file:write(tostring(AverageW))
		file:close()

		local file = io.open("/etc/plugboard/GenerationW","w+")
		file:write(tostring(GenerationW))
		file:close()

		local file = io.open("/etc/plugboard/MainsW","w+")
		file:write(tostring(MainsW))
		file:close()

		local file = io.open("/etc/plugboard/BoilerW","w+")
		file:write(tostring(BoilerW))
		file:close()


-- And the Debug Status ( 0=Off   >1=On)     WATCH OUT FOR THIS - THE FILE SIZE GETS VERY BIG, VERY QUICK.
--                      			   I PHYSICALLY COPY THE RunningW FILE TO MY PC THEN EMPTY THE
--                                           FILE ON THE HAH EVERY 24 HOURS....A LITTLE CRONTAB ACTUALLY
--                                           DOES THAT FOR ME.

		local file = io.open("/etc/plugboard/DebugStatus","r")
		DebugStatus = tonumber(file:read("*all"))
		file:close()
		if tonumber(DebugStatus) > 0 then
			local cmd = string.format(os.date("%d/%m/20%y %X")..", "..MainsW..", "..GenerationW..", "..BoilerW..",  "..ConsumedW.."  , "..EST..", "..QUAD..", "..GEN.."\n")
			local file = io.open("/etc/plugboard/RunningW","a")
			file:write(tostring(cmd))

--********* Wanted to see if Q table was filled, but never got here!!
--			local cmd = string.format(info.username..", "..Q.q1..", "..Q.q2..", "..Q.q3..",  "..Q.q4.."  , "..Q.q5.."\n")
--			file:write(tostring(cmd))

			file:close()
		end
		Post = 1
	else
		Post = Post + 1
	end

	if (MainsW > 1 and GenerationW > -1 and ConsumedW > 0) then
		RAWDataPost15()
		ESTDataPost16()
		USEDataPost17()
		BoilerDataPost18()	
	end

	self:reset()

end


-- ##############################################################
-- THESE GET THE CURRENTCOST VALUES

function GenPowerLog()
	GenPower = tonumber(xap.getValue("input.state","text"))
	TempGenerationW = tonumber(GenPower)
end

function MainsPowerLog()
	MainsPower = xap.getValue("input.state","text")
	TempMainsW= tonumber(MainsPower)
end

function BoilerPowerLog()
	BoilerPower = xap.getValue("input.state","text")
	TempBoilerW= tonumber(BoilerPower)
end


-- THESE POST THE RESULTS TO PACHUBE
-- ********************************************************************************************  CHANGE TO SUIT YOUR SETUP

function RAWDataPost15()
	CumulativeData = string.format([[xap-header
{
class=pachube.update
target=dbzoo.livebox.HAH11.pachube
}
datastream
{
id=15
tag=RAW
Value=%s
}]], MainsW)

xap.sendShort(CumulativeData)
end

function ESTDataPost16()
	CumulativeData = string.format([[xap-header
{
class=pachube.update
target=dbzoo.livebox.HAH11.pachube
}
datastream
{
id=16
tag=EST
Value=%s
}]], EST)

xap.sendShort(CumulativeData)
end

function USEDataPost17()
	CumulativeData = string.format([[xap-header
{
class=pachube.update
target=dbzoo.livebox.HAH11.pachube
}
datastream
{
id=17
tag=CALC
Value=%s
}]], ConsumedW)

xap.sendShort(CumulativeData)
end

-- BOILER POST ONLY
function BoilerDataPost18()
	CumulativeData = string.format([[xap-header
{
class=pachube.update
target=dbzoo.livebox.HAH11.pachube
}
datastream
{
id=18
tag=BoilerW
Value=%s
}]], BoilerW)

xap.sendShort(CumulativeData)
end
